// GetPrivateProfileString() -- close implementation of
// Windows NT System Services version of GetPrivateProfileString()
// doesn't provide Microsoft backwards compatability
// wrt TAB characters in the value string -- Microsoft terminates value 
// at the first TAB, but I couldn't discover what the behavior should 
// be regarding TABS in quoted strings so, I treat tabs like any other 
// characters -- NO comments following value string separated by a TAB 
// are allowed (that is an anachronism anyway)
// Added code to search for ODBC_INI file in users home directory on 
// Unix

// modified by John Sharp
// 10/15/98
// full compatability with Window's API functions except for tab char usage
// mentioned above
// and that it recognizes both unix or window's comments: #comment, ;comment     
     
// for portability
#define UNIX     1
#define stricmp(s1,s2) strcasecmp(s1,s2) 
typedef int BOOL;
     
#include <stdio.h>
     
#include <pwd.h>
     
#include <sys/types.h>
#include <string.h>
#include "profile.h"
     
#define MAXPATHLEN     255
     
// if 1st char of filename is not "/", assumes file is in login directory
void GetFullPathToFile(const char* theIniFileName,char iniFullPath[MAXPATHLEN+1]
)
{
     int j,k;
     char* ptr=0;
     struct passwd *pwd;
// since this is UNIX - absolute paths begin with "/"
     strncpy(iniFullPath,theIniFileName,MAXPATHLEN); 
     iniFullPath[MAXPATHLEN]=0;
     if (*theIniFileName=='/')
          return;
// assume file is in login home directory
     j = strlen(theIniFileName) + 1;
     pwd = getpwuid(getuid());
     if (pwd != NULL)
          ptr = pwd->pw_dir;
// if no login directory, assume /home
     if( ptr == NULL || *ptr == '\0' )
          ptr = "/home";
     k=strlen(ptr);
     if (j+k+1>MAXPATHLEN)
          *iniFullPath=0;
     else
          sprintf( iniFullPath, "%s/%s",ptr,theIniFileName );
     return;
}
     
// get PrivateProfileString
// normal usage:
//   (sectName,keyName,default,buffer,len,fileName)
//   returns length of value or length of default
//   0 if no value found or default=="" or NULL
// Otherusage returns a multistring like:
//   key1name\0key2name\0key3name\0 .... keynname\0\0
// getkeynames usage
//   (secName,NULL,NULL,buffer,len,filename)
//   returns total length of all keynames found in that section
//   zero if section not found or has no keys
// getsection names usage
//   (NULL,NULL,NULL,buffer,len,filename)
//   returns total length of all section names in the file
const char quotes[]={'\'','\"'};
int
GetPrivateProfileString(char *theSection,     // section name
     char *theKey,          // search key name
     char *theDefault,     // default value if not found 
     char *theReturnBuffer,     // return value stored here
     size_t theReturnBufferLength,     // byte length of return buffer 
     char *theIniFileName)          // pathname of ini file to search
{
     char buf[MAXPATHLEN+1];
     FILE* aFile = 0;
     size_t aLength;
     char aLine[2048];
     char *aValue;
     char *aString;
     size_t aLineLength;
     size_t aReturnLength = 0;
     
     BOOL aSectionFound = 0;
     BOOL aKeyFound = 0;
     BOOL makeMultiString;
     int j = 0;
// if no buffer to return value
     if(theReturnBufferLength == 0 || theReturnBuffer == NULL)
          return 0;
	 *theReturnBuffer='\0';
// is the return a multistring like key1\0key2\0...keyn\0\0 ? 
// if so, the return buffer must accommodate an extra null char
     if (makeMultiString=(theKey==NULL) || (theSection==NULL))
          theReturnBufferLength--;
     
     GetFullPathToFile(theIniFileName,buf); 
     aFile = fopen(buf, "r");
     
// note: if no file could be opened ...
// skip past file processing and return default
     if (aFile != NULL) while(fgets(aLine, sizeof(aLine), aFile) != NULL) 
     {
          aLineLength = strlen(aLine);
          // strip final '\n'
          if(aLineLength > 0 && aLine[aLineLength - 1] == '\n') 
          {
               aLine[aLineLength - 1] = '\0';
          }
          switch(*aLine)
          {
               case ' ': // blank line
               case ';': // comment line
               case '#': // unix comment
                    continue;
               break;
     
               case '[':     // section marker
     
                    if( aString = strchr(aLine, ']') )
                         *aString = '\0';
                // is this the section?
                    aSectionFound=theSection && aString && !stricmp(aLine + 1, theSection);
                // if saving all section names
				    if (theSection==NULL && aString)
                    {
                         aLength=strlen(aLine+1);
                         j=theReturnBufferLength-aReturnLength; 
                         if (aLength<=j)
                         {
                              strcpy(theReturnBuffer+aReturnLength,aLine+1); 
                              aReturnLength+=aLength+1;
                         }
                    }
					break;
               default:
                    // try to match value keys if in proper section
                    if(aSectionFound)
                    {
                         // try to match requested key
                         if( aValue = strchr(aLine, '=') )
                         {
                              aString=aValue-1;
                              *aValue = '\0';
                              ++aValue;
                         // strip leading blanks in value field
                              while(*aValue == ' ')
                              {
                                   *aValue++;
                              }
                         }
                         else
                              break;
                         // strip trailing blanks from key
                         while(aString >= aLine && *aString == ' ')
                              *aString-- = '\0';
                         // see if key is matched
                         if(theKey == NULL || !stricmp(theKey, aLine)) 
                         {
                              // matched -- first, terminate value part
     
                              aKeyFound = 1;
                              aLength = strlen(aValue);
                           // unquote value if quoted, but only the outer quotes
                              for (j=0;j<2;j++)
                              {
                                 if(aLength >= 2 && aValue[0] == quotes[j] &&
                                   aValue[aLength-1] == quotes[j])
                                 {
                                   aValue[aLength-1] = 0;
                                   ++aValue;
                                   aLength -= 2;
								   break;
                                 }
                              }
                           // if returning the key
						   // note: the goto can be removed and it will still work
						      if (theKey!=NULL)
                              {
                                      aLineLength=aLength<theReturnBufferLength?
											aLength: theReturnBufferLength;
                                   strncpy(theReturnBuffer,aValue,aLineLength); 
                                   aReturnLength=aLineLength;
                                   goto skip_rest_of_file;
                              }
                          // if returning all keynames
						  // only save if there is room left in the buffer
						      else  
							  {
                                 j=theReturnBufferLength-aReturnLength; 
                                 aLength=strlen(aLine);
                                 if (aLength<=j)
                                 {
                                   strcpy(theReturnBuffer+aReturnLength,
                                        aLine);
                                   aReturnLength += aLength+1;	// one extra for null
                                 }
                              }
                         }
                    }
     
               break;
          }
     }
     
skip_rest_of_file:
     if(aFile)
          fclose(aFile);
// if no key was found, return the default string
     if(!(makeMultiString || aKeyFound)) 
	 { 
          if (theDefault==NULL)
               theDefault="";
          j=strlen(theDefault);
          aReturnLength = theReturnBufferLength < j ?
               theReturnBufferLength : j;
          strncpy(theReturnBuffer, theDefault, aReturnLength);
     }
// append a trailing zero byte on the return buffer
     theReturnBuffer[aReturnLength]='\0'; 
     return aReturnLength;
}
     
     
// returns 0 if no change made to file, 1 if any change
// requires sectionname
// if keyname=NULL, removes entire section
// if valuename=NULL, removes the key
int
WritePrivateProfileString(char *theSection,     // section name
     char *theKey,          // write key name 
     char *theValue,     // input buffer
     char *theIniFileName)     // pathname of ini file to write
{
     char buf[MAXPATHLEN+1];
     FILE* aFile = 0, *tFile;
     char aLine[2048];
     char *aValue;
     char *aString;
     size_t aLineLength;
     size_t aReturnLength = 0;
     long pbegl=0, pendl=0, brepl, erepl=0; 
     BOOL retval=0;
     
     BOOL aSectionFound = 0, keyFound=0, sectFound=0; 
     int j = 0;
     
     if(theSection==NULL || theIniFileName == NULL)
          return 0;
     
     GetFullPathToFile(theIniFileName,buf);
     
     aFile = fopen(buf, "r");
// if no file, we might create one
     if(!aFile)
     {
          if (theKey==NULL || theValue==NULL)
               return 0;
     }
// use a temporary file to copy the original
     tFile=tmpfile();
     if (aFile)
     {
       while(fgets(aLine, sizeof(aLine), aFile) != NULL) 
       {
     // copy old file to temporary and keep track of where modifications
	 // might be made
	      fputs(aLine,tFile);
          pbegl=pendl;					// beginning of line
          pendl=ftell(tFile);			// beginning of next line
          aLineLength = strlen(aLine);
          // strip final '\n'
          if(aLineLength > 0 && aLine[aLineLength - 1] == '\n') 
          {
               aLine[aLineLength - 1] = '\0';
          }
          switch(*aLine)
          {
               case ' ': // blank line
               case ';': // comment line
               case '#': // unix comment
						break;
     
               case '[':     // section marker
     
                    if (aString = strchr(aLine, ']'))
                         *aString = '\0';
               // accept as matched if key exact match
                    aSectionFound=aString && stricmp(aLine + 1, theSection)==0; 
                    if (aSectionFound)
                    {
                         sectFound=1;
                   // if deleting the key, remember start of delete range
				         if (theKey==NULL)
                         {
                              brepl=pbegl;
                              erepl=pendl;
                         }
                    }
               break;
     
               default:
                    // try to match value keys if in proper section
                    if(aSectionFound)
                    {
                         // try to match requested key
     
                         if(aString = strchr(aLine, '='))
                         {
                         // strip trailing blanks from key
     
                              *aString--=0;
                              while(aString>aLine && *aString == ' ')
                                   *aString--=0;
     
                         // see if key is matched
                         // if deleting entire section, update end of delete range
						      if (theKey==NULL)
                                   erepl=pendl;
                         // if this is the key, remember where the line is
						      else if (stricmp(theKey, aLine)==0)
                              {
                                   brepl=pbegl;
                                   erepl=pendl;
                                   keyFound=1;
                              }
                         // if no match found yet in this section
						      else if (!keyFound)
                                   brepl=pendl;
                         }
                    }
               break;
          }
       }
     }
// necessary if nothing in file, so we read back at least one line
     if (!ftell(tFile))
       fputs("\n",tFile);
// close the input file     
     fclose(aFile);
// no changes if:
//  deleting a section and section not found 
//  deleting a key and key not found
     if (
          (theKey==NULL && !sectFound)
          || (theKey && theValue==NULL && !keyFound)
        )
        goto no_changes;
// no changes if can't open ini file for output
     aFile=fopen(buf,"w");
     if (aFile==NULL)
        goto no_changes;
// read back the temp file and update the ini file
     retval=1;
     rewind(tFile);
     pendl=0;
     while (fgets(aLine,sizeof(aLine),tFile)) 
     {
          pbegl=pendl;
          pendl=ftell(tFile);
     // if deleting a section
          if (theKey==NULL && pbegl>=brepl && pendl<=erepl)
               continue;
     // found line that matches the key
          if (theKey!=NULL && pbegl==brepl)
          {
          // if updating with a new value
               if (theValue!=NULL)
               {
                    if (keyFound)
                         sprintf(aLine,"%s = %s\n",
                              theKey,theValue);
                    else
                         sprintf(aLine,"[%s]\n%s= %s\n",
                              theSection,theKey,theValue);
               }
      // if deleting the key
               else if (keyFound)
                    continue;
          }
      // if copying this line of the tempfile
	      fputs(aLine,aFile);
     }
// close output file if any
     fclose(aFile);
no_changes:
// and close tempfile
     fclose(tFile);
     return retval;
}
